# -*- coding: utf-8 -*-
"""
File Templates
"""

import pathlib
from types import SimpleNamespace

SETTINGS_TEXT = '''
# -*- coding: utf-8 -*-
"""
{ Settings }
"""

import pathlib

# Base Directory
BASE_DIR = pathlib.Path(__file__).parents[1]

# Installed Apps
INSTALLED_APPS = []
'''.strip()

CONFIG_TEXT = '''
# -*- coding: utf-8 -*-
"""
{ Config }
"""
'''.strip()


SPOC_TEXT = """
[spoc]
# options: production, development, staging
mode = "development"

# Modes
[spoc.apps]
production = []
development = []
staging = []

# Extras
[spoc.extras]
""".strip()


ENV_TEXT = """
[env] # Environment Settings
""".strip()


def start_project():
    """Create Core Files"""

    # import __main__

    # root_dir = pathlib.Path(__main__.__file__).parent
    root_dir = pathlib.Path.cwd()

    path = SimpleNamespace()

    path.config_dir = root_dir / "config"
    path.config_file = path.config_dir / "__init__.py"
    path.settings_file = path.config_dir / "settings.py"
    path.spoc_file = path.config_dir / "spoc.toml"

    # Create Folder
    path.config_dir.mkdir(exist_ok=True)

    # Environment Folder
    env_types = ["production", "development", "staging"]
    env_path = path.config_dir / ".env"
    env_path.mkdir(exist_ok=True)

    # Create Files
    if not path.config_file.exists():
        with open(path.config_file, "w", encoding="utf-8") as file:
            file.write(CONFIG_TEXT)

    if not path.settings_file.exists():
        with open(path.settings_file, "w", encoding="utf-8") as file:
            file.write(SETTINGS_TEXT)

    if not path.spoc_file.exists():
        with open(path.spoc_file, "w", encoding="utf-8") as file:
            file.write(SPOC_TEXT)

    # Environment Create Files
    for env in env_types:
        current_path = env_path / f"{env}.toml"
        if not current_path.exists():
            with open(current_path, "w", encoding="utf-8") as file:
                file.write(ENV_TEXT)



"""
Exception & Error Messages
"""

from typing import Any, TypeVar

T = TypeVar("T", bound="FrozenDict")


class FrozenDict(dict):
    """Immutable Dictionary"""

    def fromkeys(self, key, value):  # type:ignore
        """
        Create a new FrozenDict with keys from the given iterable and set to the provided value.
        """
        return type(self)(dict(self).fromkeys(key, value))

    def __getattribute__(self, attribute: Any):
        """
        Prevent modification methods from being accessed.
        """
        if attribute in ("clear", "update", "pop", "popitem", "setdefault"):
            raise AttributeError(
                f"{type(self).__name__} object has no attribute {attribute}"
            )
        return dict.__getattribute__(self, attribute)

    def __setitem__(self, key: Any, value: Any):
        """
        Prevent item assignment.
        """
        raise TypeError(
            f"{type(self).__name__} object does not support item assignment"
        )

    def __delitem__(self, key: Any):
        """
        Prevent item deletion.
        """
        raise TypeError(f"{type(self).__name__} object does not support item deletion")

    def __hash__(self):  # type:ignore
        """
        Compute the hash value of the FrozenDict.
        """
        return hash(tuple(sorted(self.iteritems())))

    def __repr__(self) -> str:
        """
        Return the string representation of the FrozenDict.
        """
        return f"{self.__class__.__name__}({super().__repr__()})"



# -*- coding: utf-8 -*-
"""
Elastic Framework Builder
"""

from .installer import start_project

try:
    # Core Tools
    from .components import component, is_component

    # Frame Work
    from .framework import Spoc as App
    from .importer import frozendict
    from .importer.base import search_method as search
    from .importer.tools import get_fields
    from .singleton import singleton

    # Globals
    base_dir = App.base_dir
    config = App.config
    mode = App.mode
    env = App.environment
    settings = App.settings
except ImportError:
    pass

__all__ = (
    # Tools
    "start_project",
    "component",
    "frozendict",
    "get_fields",
    "is_component",
    "search",
    "singleton",
    # Globals
    "base_dir",
    "config",
    "mode",
    "settings",
)



# -*- coding: utf-8 -*-
"""
Tool for Singleton(s)
"""


class Singleton:
    """Class: Create a Singleton"""

    def __new__(cls, *args, **kwargs):
        it_id = "__it__"
        it = cls.__dict__.get(it_id, None)
        if it is not None:
            return it
        it = object.__new__(cls)
        setattr(cls, it_id, it)
        it.init(*args, **kwargs)
        return it

    def init(self, *args, **kwargs):
        """Class __init__ Replacement"""


def singleton(cls):
    """Function: Create a Singleton"""
    custom_class = type(
        cls.__name__,
        (cls, Singleton),
        {},
    )
    return custom_class



# -*- coding: utf-8 -*-
"""
FrameWork
"""

import os
import pathlib
from typing import Any, Dict, List, Optional

from .importer import create_framework, frozendict
from .inject import (
    collect_extras,
    collect_installed_apps,
    get_app_mode,
    get_toml_files,
    inject_apps_folder,
    load_envs,
)
from .singleton import singleton

# Attempt to import the configuration module, create base if it does not exist.
CONFIG = None
try:
    import config  # type: ignore

    CONFIG = config
except ImportError:
    CONFIG = None

if CONFIG:
    try:
        from config import settings  # type: ignore
    except ImportError as exception:
        raise ValueError("Missing { ./config/__init__.py } module.") from exception

    if not settings:
        raise ValueError("Missing { ./config/settings.py } module.")

    if not hasattr(settings, "BASE_DIR"):
        raise ValueError("Missing { BASE_DIR } in file { ./config/settings.py }.")

    # Core configuration and settings
    SETTINGS = settings
    BASE_DIR = settings.BASE_DIR

    # Inject the applications folder to the global namespace
    inject_apps_folder(BASE_DIR)

    # Load and parse configuration files
    TOML_DIR = get_toml_files(BASE_DIR)
    MODE = get_app_mode(TOML_DIR)
    EXTRAS = TOML_DIR["spoc"].get("spoc", {}).get("extras", {})

    # Load environment variables
    TOML_DIR["env"] = load_envs(BASE_DIR, MODE)
    TOML_DIR = frozendict(TOML_DIR)

    @singleton
    class Spoc:
        """Core Object (Singleton) for the whole Framework."""

        # Core
        base_dir: pathlib.Path = BASE_DIR
        config: Dict[str, Dict] = TOML_DIR
        mode: str = MODE
        settings: Any = SETTINGS
        environment: Any = TOML_DIR.get("env", {})

        # Python Modules
        module: Any = None

        # App
        installed_apps: List | None = None
        extras: Dict[Any, Any] | None = None
        component: Dict[Any, Any] | None = None

        def init(self, plugins: Optional[List] = None) -> None:
            """
            Initialize the framework by collecting installed applications and extras.

            Args:
                plugins (Optional[List]): A list of plugins to initialize with the framework.
            """
            # GLOBALS
            installed_apps = collect_installed_apps(TOML_DIR, SETTINGS)
            extras = collect_extras(EXTRAS)

            plugins = plugins or []
            framework = create_framework(
                plugins=plugins, installed_apps=installed_apps, extras=extras
            )

            # Output Model
            self.installed_apps = installed_apps
            self.extras = framework.extras
            self.module = framework.module
            self.component = framework.plugin

            os.chdir(BASE_DIR)

        @staticmethod
        def keys() -> List[str]:
            """
            Get the list of keys for the framework object.

            Returns:
                List[str]: A list of attribute keys.
            """
            return [
                "base_dir",
                "config",
                "installed_apps",
                "mode",
                "project",
                "settings",
                "extras",
                "module",
                "component",
            ]



# -*- coding: utf-8 -*-
"""[ Collect & Inject ]
Project (Module)
"""

import os
import pathlib
import sys
from typing import Any

from .importer import frozendict
from .importer.base import search_method
from .toml_core import TOML


def inject_apps_folder(base_dir: pathlib.Path):
    """Inject { ./apps } Directory"""

    base_apps = pathlib.Path(base_dir / "apps")
    base_apps.mkdir(parents=True, exist_ok=True)
    if base_apps.exists():
        sys.path.insert(0, os.path.join(base_dir, "apps"))


def collect_apps_partial(app_mode: str, the_apps: dict):
    """Collect All Apps"""
    installed_apps = []
    match app_mode:
        case "production":
            installed_apps.extend(the_apps.get("production", []))
        case "staging":  # production + staging
            installed_apps.extend(the_apps.get("production", []))
            installed_apps.extend(the_apps.get("staging", []))
        case "development":  # production + staging + development
            installed_apps.extend(the_apps.get("production", []))
            installed_apps.extend(the_apps.get("staging", []))
            installed_apps.extend(the_apps.get("development", []))
    return installed_apps


def collect_installed_apps(toml_dir, settings: Any = None):
    """Collect All Apps"""

    # Step[1]: INIT { Values }
    toml_spoc = toml_dir["spoc"].get("spoc", {})
    the_apps = toml_spoc.get("apps", {})
    app_mode = toml_spoc.get("mode", "development")

    # Installed Apps
    installed_apps = []

    # Step[2]: Collect Apps
    installed_apps.extend(collect_apps_partial(app_mode, the_apps))
    if hasattr(settings, "INSTALLED_APPS"):
        installed_apps.extend(settings.INSTALLED_APPS)

    return list(set(installed_apps))


def get_toml_file(toml_file: pathlib.Path):
    """Get { TOML } File"""

    manager = TOML(toml_file)

    if toml_file.exists():
        toml_dict = frozendict(**manager.read())
    else:
        toml_dict = frozendict({})
    return toml_dict


def get_toml_files(base_dir: pathlib.Path):
    """Collect All { TOML } Files"""

    return {
        "spoc": get_toml_file(base_dir / "config" / "spoc.toml"),
        "pyproject": get_toml_file(base_dir / "pyproject.toml"),
    }


def get_app_mode(toml_dir):
    """Get App { Mode }"""
    return toml_dir["spoc"].get("mode", "development")


def collect_extras(extras):
    """Collect All Extra { Python-Objects }"""

    all_extras_modules = {}
    for group, items in extras.items():
        all_extras_modules[group] = []
        for object_uri in items:
            obj = search_method(object_uri)
            if obj:
                all_extras_modules[group].append(obj)
    return frozendict(all_extras_modules)


def load_envs(base_dir, mode):
    """Load Environment Variables"""
    file_path = base_dir / "config" / ".env" / f"{ mode }.toml"
    return get_toml_file(file_path).get("env", {})



# -*- coding: utf-8 -*-

"""
Tool for handling TOML files
"""

import tomllib
from pathlib import Path
from typing import Any, Dict


class TOML:
    """A wrapper class for managing TOML files."""

    def __init__(self, file: Path):
        """
        Initialize the TOML manager with the given file path.
        """
        self.file = file

    def read(self) -> Dict[str, Any]:
        """
        Read and parse the TOML file.
        """
        with open(self.file, "rb") as active_file:
            parsed_toml = tomllib.load(active_file)
        return parsed_toml



# -*- coding: utf-8 -*-

"""
Module (Core)
"""

from typing import Any

from .base import get_plugins
from .frozendict import FrozenDict as frozendict
from .get_spoc import get_spoc
from .types import App


def global_dict(core) -> dict:
    """Create Global-Dict"""

    all_modules_dir = {}
    for variables in core.plugins.values():
        for setup in variables:
            base_uri = f"{setup.app}.{setup.module}"
            for key, field in setup.fields.items():
                if not key == "spoc":
                    module_uri = base_uri + f".{key}"
                    all_modules_dir[module_uri] = field
    return frozendict(all_modules_dir)


def create_framework(
    plugins: list[str],
    installed_apps: list[str],
    extras: dict[str, Any] | None = None,
) -> App:
    """Create Framework"""

    core = get_plugins(plugins, installed_apps)
    spoc = get_spoc(core.plugins)
    python_variables = global_dict(core)
    return App(
        extras=extras,
        plugin=spoc.schema,
        module=python_variables,
    )



# -*- coding: utf-8 -*-

"""
DataClasses

* Class
* Global
* Core
* Definition
* App
"""

import dataclasses as dc
import typing


@dc.dataclass(frozen=True)
class Class:
    """Project Classes"""

    name: str | None = None
    app: str | None = None
    module: str | None = None
    key: str | None = None
    uri: str | None = None
    object: typing.Any = None


@dc.dataclass(frozen=True)
class Spoc:
    """Project Spoc (Globals)"""

    modules: typing.Any = None
    schema: typing.Any = None


@dc.dataclass(frozen=True)
class Core:
    """Project Core"""

    modules: dict
    plugins: dict


@dc.dataclass(frozen=True)
class Definition:
    """Project Definitions"""

    path: str
    app: str
    module: str
    fields: dict[str, typing.Any]


@dc.dataclass(frozen=True)
class App:
    """Project App"""

    plugin: typing.Any
    module: dict[str, typing.Any]
    extras: dict[str, typing.Any] | None = None
